#ifndef __DAS_INFO__HPP__
#define __DAS_INFO__HPP__

#include <string>
#include <cmath>
using namespace std;

#include "Logsystem.hpp"

extern "C"{
#include <stdint.h>
}

/// @class DAS_Info 
/// A Class for essential parameters of Data Acquisition System
/// Default Parameter is 16Msps/2bit/4ch.
class DAS_Info{
protected:
  int32_t srate; ///< Sampling rate (Hz)
  int      nch;  ///< Nch
  int      qbit; ///< Quantization bits;
  int      first_ch; ///< First Channel number, starting from 0.
public:
  /// Constructor 
  DAS_Info():srate(4000000),nch(1),qbit(1),first_ch(0){;}
  /// Constructor 
  DAS_Info(const uint32_t sr, const int nc=4, const int q=2, const int fch=0):
    srate(sr),nch(nc),qbit(q), first_ch(fch){};
  int32_t Srate() const{return srate;} ///< Sampling Rate (Hz).
  int  Nch() const {return nch;}    ///< Nch Number of channels
  int  Qbit() const {return qbit;}  ///< Quantization bits.
  int  FirstCh()const {return first_ch;} ///< Returns First channel 0,1,...
  /** Settting the Sampling Parameters. sr (Hz)*/
  void Set_Param(const int32_t nc, const int32_t q,const uint32_t sr){
    srate=sr; nch=nc; qbit=q;}
  /** Settting the Sampling rate in Hz. */
  void Set_Srate(const uint32_t Hz){srate=Hz;}
  /** Settting the Number of channels . */
  void Set_Nch(const int n){nch=n;}
  /** Settting the Quantization number of bit. */
  void Set_Qbit(const int q){qbit=q;}
  /** Settting the first channel starting from 0. */
  void SetFirstCh(const int n){first_ch=n;}
  /** Returns Data Rate in Bytes per second. */
  uint32_t Bytes_sec()const {return (srate/8*nch*qbit);} /// Bytes/sec.
  /** Returns sampling mode in string [xxMsps/xbit/xch] */
  string Mode(const int fmt=1)const; 
  /** Returns Total data rate [Words(4byte)/sec] */
  int32_t Speed()const {return (srate/32*qbit*nch);} 
  /**  Returns mode subset  test */
  const bool IsSubset(const DAS_Info& das) const;
  /** Splitting mode input = xxMsps/xbit/xxch */
  const bool SetMode(const string & mode);
};

inline const bool DAS_Info::IsSubset(const DAS_Info& das) const{
  if(das.Nch() > nch || das.Qbit() > qbit || das.Srate() > srate ){
    Logsystem::error_logger()->Tstamp()
      << "E DAS_Info::IsSubset(): mode(" << das.Mode() << ") is not subset of OrgMode ("
      << Mode() << ")\n"; 
    return false;
  }
  if(das.Qbit() > qbit){
    Logsystem::error_logger()->Tstamp()
      << "E DAS_Info::IsSubset(): Qbit is in the range of subset, "
      << "though it is not implemented yet. Large(" << Mode() << ") > small("
      << das.Mode() <<")\n";
    return false;
  }
  
//   if(das.Srate() > srate){
//     Logsystem::error_logger()->Tstamp()
//       << "E  DAS_Info::IsSubset(): Srate is in the range of subset, "
//       << "though under-sampling is not implemented yet. Large(" 
//       << Mode() << ") > small(" << das.Mode() <<")\n";
//     return false;
//   }

  return true;
}

inline  string DAS_Info::Mode(const int fmt)const{
  char  buf[256];
  switch (fmt){
  case 0:
    sprintf(buf,"%02d:%d:%02d", 
	    (int)rint((double)srate/1000000),qbit,nch);
    break;
  case 1:
    sprintf(buf,"%02dMsps:%dbit:%02dch",
	    (int)rint((double)srate/1000000),qbit,nch);
    break;
  case 2:
    sprintf(buf,"%02dMsps/%dbit/%02dch",
	    (int)rint((double)srate/1000000),qbit,nch);
    break;
  case 3:
    sprintf(buf,"%02dMsps/%dbit/%02dch:%d",
	    (int)rint((double)srate/1000000),qbit,nch, first_ch);
    break;

  default:
    break;
  }
  string ret = buf;
  return ret;
}

/** Splitting mode input = xxMsps/xbit/xxch */
inline const bool DAS_Info::SetMode(const string& mode){

  try{
    if(mode.find("Msps") != string::npos){  // xxMsps/xbit/xxch format  
      int pos = mode.find("Msps/"); 
      if(pos < 0) throw  "Input mdoe missing 'Msps/'";
      string work = mode.substr(pos+5);
      int Msps = atoi(mode.substr(0,pos).c_str());
      pos = work.find("bit/"); 
      if(pos < 0) throw  "Input mdoe missing 'bit/'";
      int bit = atoi(work.substr(0,pos).c_str());
      work = work.substr(pos+4); 
      pos = work.find("ch");
      if(pos < 0) throw  "Input mdoe missing 'ch'";
      int nch = atoi(work.substr(0,pos).c_str());
      Set_Param(nch, bit, Msps*1000000);  
    }else if(mode.find(":") != string::npos){
      int ipos = mode.find(":");
      int ipos2 = mode.find(":",ipos+1);
      int Msps = atoi(mode.substr(0,ipos).c_str());
      int bit = atoi(mode.substr(ipos+1,ipos2-ipos-1).c_str());
      int nch = atoi(mode.substr(ipos2+1).c_str());
      Set_Param(nch, bit, Msps*1000000);  
    }else{
      throw "E DAS_Info::SetMode(): mode-format is not xxMsps/xbit/xxch nor xx:xx:xx";
    }
    fprintf(stderr,"I DAS_Info::SetMode(): %2s is set.\n", Mode().c_str());
    return true;

  }

  catch(const char* msg){
    cerr << msg << endl;
    return false;
  }
  

}


#endif
